#include <strstream>
#include <stdexcept>
#include <cassert>
#include <iomanip>

#include "currency.h"

// For whatever reason C++Builder bitchs about conversions losing significant
// digits even in cases such as:
//  char x ;
//  x /= 10 ;
//  x = y % 10 ;
#pragma warn -sig
// Disable unsigned/signed comparison
#pragma warn -csu

using namespace std ;

// Formatting definitions

const char DECIMALCHARACTER = '.' ;   // Decimal Point Symbol
const char COMMACHARACTER = ',' ;     // Digit Grouping Symbol
const char ERRORCHARACTER = '#' ;     // Formatting Error Character
const int DEFAULTDECIMALPLACES = 2 ;  // Default number of display decimal places

// ios xalloc indices
int currency::whole_index ;     // Number of whole digits
int currency::decimal_index ;   // Number of decimal places
int currency::negative_index ;  // Negative format
int currency::fill_index ;      // Fill Character
int currency::flags_index ;     // Formatting flags

const int SIGN_FLAG = 1 ;       // Always show sign flag
const int COMMA_FLAG = 2 ;      // Use comma separators

//
//  Description:
//
//    The CurrencyInitialization class performs the initialization
//    required by the currency class. The definitionis here to
//    ensure that only one object gets created. The object is defined
//    in the unnamed namespace below.
//
//    The class constructor allocates the indices for the
//    formatting settings used by the currency class.
//
class CurrencyInitialization
{
public:
  CurrencyInitialization ()
  {
    currency::whole_index = ios::xalloc () ;
    currency::decimal_index = ios::xalloc () ;
    currency::negative_index = ios::xalloc () ;
    currency::fill_index = ios::xalloc () ;
    currency::flags_index = ios::xalloc () ;
  }
}  ;

namespace
{
CurrencyInitialization InitializationObject ;

const char *numbers [100] =
{
"", "One", "Two", "Three", "Four",
"Five", "Six", "Seven", "Eight", "Nine",
"Ten", "Eleven", "Twelve", "Thirteen", "Fourteen",
"Fifteen", "Sixteen", "Seventeen", "Eighteen", "Nineteen",
"Twenty", "Twenty-One", "Twenty-Two",  "Twenty-Three",  "Twenty-Four",
"Twenty-Five", "Twenty-Six", "Twenty-Seven",  "Twenty-Eight",  "Twenty-Nine",
"Thirty", "Thirty-One", "Thirty-Two",  "Thirty-Three",  "Thirty-Four",
"Thirty-Five", "Thirty-Six", "Thirty-Seven",  "Thirty-Eight",  "Thirty-Nine",
"Forty", "Forty-One", "Forty-Two",  "Forty-Three",  "Forty-Four",
"Forty-Five", "Forty-Six", "Forty-Seven",  "Forty-Eight",  "Forty-Nine",
"Fifty", "Fifty-One", "Fifty-Two",  "Fifty-Three",  "Fifty-Four",
"Fifty-Five", "Fifty-Six", "Fifty-Seven",  "Fifty-Eight",  "Fifty-Nine",
"Sixty", "Sixty-One", "Sixty-Two",  "Sixty-Three",  "Sixty-Four",
"Sixty-Five", "Sixty-Six", "Sixty-Seven",  "Sixty-Eight",  "Sixty-Nine",
"Seventy", "Seventy-One", "Seventy-Two",  "Seventy-Three",  "Seventy-Four",
"Seventy-Five", "Seventy-Six", "Seventy-Seven",  "Seventy-Eight",  "Seventy-Nine",
"Eighty", "Eighty-One", "Eighty-Two",  "Eighty-Three",  "Eighty-Four",
"Eighty-Five", "Eighty-Six", "Eighty-Seven",  "Eighty-Eight",  "Eighty-Nine",
"Ninety", "Ninety-One", "Ninety-Two",  "Ninety-Three",  "Ninety-Four",
"Ninety-Five", "Ninety-Six", "Ninety-Seven",  "Ninety-Eight",  "Ninety-Nine",
} ;

}

//
//  Description
//
//    Default/Integer Conversion constructor
//
currency::currency (int value)
{
  memset (data, 0, sizeof (data)) ;

  // Convert the digits to text.
  strstream strm ;
  if (value >= 0)
  {
    strm << value ;
    sign = 1 ;
  }
  else if (value < 0)
  {
    strm << - value ;
    sign = -1 ;
  }

  // Store the digits in the object.
  char *buffer = strm.str () ;
  int len = strlen (buffer) ;
  if (len > WHOLEDIGITS)
    throw std::overflow_error ("value too large") ;

  int pos = DECIMALPLACES ;
  for (int ii = len - 1 ; ii >= 0 ; -- ii, ++ pos)
  {
    data [pos] = (buffer [ii]  - '0') ;
  }

  CheckState () ;
  return ;
} ;

//
//  Description:
//
//    double conversion/Constructor
//
currency::currency (double value)
{
  memset (data, 0, sizeof (data)) ;

  // Scale the double so that all the fractional digits we need
  // are to the left of the decimal point.
  double temp = value ;
  for (unsigned int ii = 0 ; ii < DECIMALPLACES ; ++ ii)
    temp *= 10.0 ;

  // Convert to string.
  strstream strm ;
  if (value > 0)
  {
    temp += 0.5 ;
    strm << setiosflags (ios::fixed) << temp ;
    sign = 1 ;
  }
  else if (value < 0)
  {
    temp -= 0.5 ;
    strm << setiosflags (ios::fixed) << - temp ;
    sign = -1 ;
  }
  else
  {
    sign = 1 ;
    strm << setiosflags (ios::fixed) << temp ;
  }

  // Copy the digits from the buffer to the object.

  char *buffer = strm.str () ;
  int len = 0 ;
  for ( ; buffer [len] != DECIMALCHARACTER && buffer [len] != '\000' ; ++ len)
    ;

  if (len > PRECISION)
    throw std::overflow_error ("value too large") ;

  for (int ii = 0 ; ii < len ; ++ ii)
    data [len - 1 - ii] =  (buffer [ii] - '0') ;

  CheckState () ;
  return ;
} ;


currency::currency (const std::string &str)
{
  memset (data, 0, sizeof (data)) ;
  int offset = 0 ;
  while (isspace (str [offset]))
    ++ offset ;

  if (str [offset] == '-')
  {
    sign = -1 ;
    ++ offset ;
  }
  else if (str [offset] == '+')
  {
    sign = 1 ;
    ++ offset ;
  }
  else
  {
    sign = 1 ;
  }

  if (offset >= str.length () || ! isdigit (str [offset]))
    return ;

  int wholeend = offset + 1 ;
  while (isdigit (str [wholeend]))
    ++ wholeend ;

  if (wholeend - offset > WHOLEDIGITS)
    throw std::overflow_error ("value too large") ;

  for (int ii = DECIMALPLACES, jj = wholeend - 1 ; jj >= offset ; ++ ii, -- jj)
    data [ii]  =  (str [jj] - '0') ;

  if (str [wholeend] == DECIMALCHARACTER)
  {
    for (int ii = DECIMALPLACES - 1, jj = wholeend + 1 ;
         isdigit (str [jj]) && ii >= 0 ;
         -- ii, ++ jj)
    {
      data [ii]  =  (str [jj] - '0') ;
    }
  }

  CheckState () ;
  return ;
}

//
//  Description:
//
//    Copy Constructor
//
currency::currency (const currency &src)
{
  DoCopy (src) ;
  CheckState () ;
  return ;
}

//
//  Description:
//
//    Assignment Operator
//
currency &currency::operator=(const currency &src)
{
  DoCopy (src) ;
  CheckState () ;
  return *this ;
}

//
//  Description:
//
//    This function implements an object copy.
//
//  Parameters:
//
//    src:  The object to copy
//
void currency::DoCopy (const currency &src)
{
  sign = src.sign ;
  for (unsigned int ii = 0 ; ii < PRECISION ; ++ ii)
    data [ii] = src.data [ii] ;
  return ;
}

//
//  Description:
//
//    This function writes an object to an output stream according to the
//    settings specified for the stream.
//
//    The only standard setting used for formatting is width()/setw().
//
//    If the currency value cannot be written correctly using the
//    current formatting settings all the digits are replaced with
//    the character ERRORCHARACTER so that the user can quickly
//    identify the problem.
//
//  Parameters:
//
//    ostream:  The output stream.
//
//  Implicit Inputs:
//
//    These member variables contain indices for the ios.iword () function
//    that hold currency-specific formatting information.
//
//      whole_index
//      decimal_index
//      negative_index
//      fill_index
//      flags_index
//

//
void currency::Print (ostream &strm) const
{
#if ! defined (SIMPLEFORMATTING)
  bool error = false ;

  // Fetch formatting settings rom the ostream object.
  int decimalplaces = strm.iword (decimal_index) ;
  int negativeformat = strm.iword (negative_index) ;
  bool showsign = (strm.iword (flags_index) & SIGN_FLAG) != 0 ;
  int wholedigits = strm.iword (whole_index) ;
  char fillcharacter = (char) strm.iword (fill_index) ;
  bool showcommas = (strm.iword (flags_index) & COMMA_FLAG) != 0 ;

  // If the number of whole digits is not specified it is
  // taken to mean the default. We allow zero decimal places
  // but not zero whole digits.
  if (wholedigits == 0)
    decimalplaces = DEFAULTDECIMALPLACES ;
  // Round to the nearest displayable value.
  currency value = Round (decimalplaces) ;

  int wholeneeded ; // The number of whole digits needed to display the value.

  // Determine the number of digits required to hold the whole
  // number part of the value.
  for (wholeneeded = PRECISION - 1 ;
       wholeneeded >= DECIMALPLACES ;
       -- wholeneeded)
  {
    if (value.data [wholeneeded] != 0)
      break ;
  }
  wholeneeded -= DECIMALPLACES - 1 ;
  // We always have at least one whole digit.
  if (wholeneeded == 0)
    wholeneeded = 1 ;

  // If the user has specified a maximum number of of digits
  // that is fewer that the number of digits required to display
  // the value then we mark it as an error.
  if (wholedigits != 0
      && wholeneeded > wholedigits)
  {
    wholeneeded = wholedigits ;
    error = true ;
  }

  // Determine the number of fill bytes.
  int fillneeded ;
  if (wholedigits == 0
      || fillcharacter == 0
      || fillcharacter == ' ')
  {
    fillneeded = 0 ;
  }
  else
  {
    fillneeded = wholedigits - wholeneeded ;
  }

  int length = 0 ;
  if (decimalplaces != 0)
    length += decimalplaces + 1 ;
  if (showcommas)
    length += (wholeneeded + fillneeded) / 3 ;
  length += wholeneeded + fillneeded ;

  if (negativeformat == paren)
  {
    // Parentheses are used to enclose negative numbers. We allocate
    // space for them even if the value is positive so that positive
    // and negative numbers will be correctly aligned.
    length += 2 ;
  }
  else if (negativeformat == trailing || showsign || value.sign < 0)
  {
    // For traiing signs we always allocate space for the sign so that
    // positive and negative numbers will be aligned.

    // If the showsign flag is set or the value is negative we need
    // space for the sign as well.
    ++ length ;
  }

  // Ensure that the value can be written in the field size given to us.
  if (strm.width () != 0 && strm.width () < length)
    error = true ;

  // At this point we have figure out all the lengths. Now we write the
  // value to a string buffer.
  string buffer ;
  buffer.resize (length) ;

  // Since numbers are stored internally in the reverse order that they
  // are displayed we use a descending counter for the output position.
  int position = length - 1 ;

  // Negative sign.
  if (negativeformat == paren)
  {
    if (value.sign < 0)
    {
      buffer [0] = '(' ;
      buffer [position] = ')' ;
      -- position ;
    }
  }
  else if (negativeformat == trailing)
  {
    if (value.sign < 0)
    {
      buffer [position] = '-' ;
    }
    else if (showsign)
    {
      buffer [position] = '+' ;
    }
    -- position ;
  }
  else
  {
    if (value.sign < 0)
    {
      buffer [0] = '-' ;
    }
    else if (showsign)
    {
      buffer [0] = '+' ;
    }
  }

  // Fractional digits (Decimal places)
  if (decimalplaces != 0)
  {
    for (int ii = DECIMALPLACES - decimalplaces ; ii < DECIMALPLACES ; ++ ii)
    {
      buffer [position] =  (value.data [ii] + '0') ; -- position ;
    }
    buffer [position] = '.' ; -- position ;
  }

  if (showcommas)
  {
    for (int ii = 0, src = DECIMALPLACES ;
         ii < wholeneeded + fillneeded ;
         ++ ii, ++ src)
    {
      if (ii < wholeneeded)
      {
        buffer [position] =  (value.data [src] + '0') ; -- position ;
      }
      else
      {
        buffer [position] = fillcharacter ; -- position ;
      }

      if ((ii + 1) % 3 == 0)
      {
        if (ii < wholeneeded - 1)
          buffer [position] = COMMACHARACTER ;
        else
          buffer [position] = fillcharacter ;
        -- position ;
      }
    }
  }
  else
  {
    for (int ii = 0, src = DECIMALPLACES ;
         ii < wholeneeded + fillneeded ;
         ++ ii, ++ src)
    {
      if (ii < wholeneeded)
      {
        buffer [position] =  (value.data [src] + '0') ; -- position ;
      }
      else
      {
        buffer [position] = fillcharacter ; -- position ;
      }
    }
  }

  // If we can't fit the correct value in the field we replace
  // all of the digits with another character so that there will
  // be no mistaking the problem.
  if (error)
  {
    for (int ii = 0 ; ii < length ; ++ ii)
    {
      if (isdigit (buffer [ii]))
      {
          buffer [ii] = ERRORCHARACTER ;
      }
    }
  }

  // Adjust to fit the field width.
  if (strm.width () != 0)
  {
    if (strm.width () > buffer.length ())
    {
      // Right justify to fit the strm field width.
      string pad ;
      pad.resize (strm.width () - buffer.length ()) ;
      buffer = pad + buffer ;
    }
    else if (strm.width () < buffer.length ())
    {
      // The value will not fit the field so we truncate from the
      // left. The error code will have already executed.

      for (int ii = 0, jj = buffer.length () - strm.width () ;
           ii < strm.width () ;
           ++ ii, ++ jj)
      {
        buffer [ii] = buffer [jj] ;
      }
      buffer.resize (strm.width ()) ;
    }
  }
  strm << buffer ;
  strm.width (0) ; // follow convention and reset the width.
#else
  if (sign < 0)
    strm << '-' ;

  bool nonzero = false ;
  for (int ii = PRECISION - 1 ; ii >= DECIMALPLACES ; -- ii)
  {
    if (data [ii] != 0 || nonzero)
    {
      strm << (char) (data [ii] + '0') ;
      nonzero = true ;
    }
  }
  if (! nonzero)
    strm << '0' ;

  strm << '.' ;
  for (int ii = DECIMALPLACES - 1 ; ii >= 0 ; -- ii)
  {
    strm << (char) (data [ii] + '0') ;
  }
#endif
  return ;
}

//
//  Description:
//
//    This function performs a signed comparison to another object.
//
//  Parameters:
//
//    op:  The object to compare to.
//
//  Return Value:
//
//    1 => This object's value is greater than that of 'op'
//    0 => The objects have the same value.
//    -1 => This object value is less than that of 'op'
//

int currency::Compare (const currency &op) const
{
  if (sign < op.sign)
    return -1 ;
  else if (sign > op.sign)
    return 1 ;
  else
  {
    if (sign >= 0)
      return CompareAbs (op) ;
    else
      return - CompareAbs (op) ;
  }
}

//
//  Description:
//
//    This function performs an unsigned comparison to another object.
//
//  Parameters:
//
//    op:  The object to compare to.
//
//  Return Value:
//
//    1 => This object's value is greater than that of 'op'
//    0 => The objects have the same value.
//    -1 => This object value is less than that of 'op'
//
int currency::CompareAbs (const currency &op) const
{
  for (int ii = REGISTERSIZE - 1 ; ii >= 0 ; -- ii)
  {
    if (data [ii] < op.data [ii])
      return -1 ;
    else if (data [ii] > op.data [ii])
      return 1 ;
  }
  return 0 ;
}

//
//  Description:
//
//    Comparison operators (<,<=,>,>=,==,!=)
//
//  Parameters:
//
//    op:  The object to compare to
//
//  Return Value:
//
//    true => Comparision is true
//    false => Comparions is false
//
bool currency::operator<(const currency &op) const
{
  if (Compare (op) < 0)
    return true ;
  else
    return false ;
}

bool currency::operator<=(const currency &op) const
{
  if (Compare (op) <= 0)
    return true ;
  else
    return false ;
}

bool currency::operator>(const currency &op) const
{
  if (Compare (op) > 0)
    return true ;
  else
    return false ;
}

bool currency::operator>=(const currency &op) const
{
  if (Compare (op) >= 0)
    return true ;
  else
    return false ;
}

bool currency::operator==(const currency &op) const
{
  if (Compare (op) == 0)
    return true ;
  else
    return false ;
}

bool currency::operator!=(const currency &op) const
{
  if (Compare (op) != 0)
    return true ;
  else
    return false ;
}

//
//  Description:
//
//    Unary minus (negation) operator
//
//  Return Value:
//
//    The negative of the object's value.
//
currency currency::operator-() const
{
  currency result (*this) ;
  result.sign = - sign ;
  result.CheckState () ;
  return result ;
}

//
//  Description:
//
//    Arithmetic binary operators (+, -, * /) and
//    arithmetic assignment operators (+=, -=, *=, /=)
//
//  Parameters:
//
//    op: The other operand
//
//  Return value:
//
//    *this OPERATOR op
//
currency currency::operator+(const currency &op) const
{
  currency result ;
  if (sign == op.sign)
  {
    result.AddUnsigned (*this, op) ;
    result.sign = sign ;
  }
  else
  {
    if (CompareAbs (op) >= 0)
    {
       result.sign = sign ;
       result.SubtractUnsigned (*this, op) ;
    }
    else
    {
       result.sign = op.sign ;
       result.SubtractUnsigned (op, *this) ;
    }
  }
  result.CheckState () ;
  return result ;
}

currency currency::operator-(const currency &op) const
{
  currency result ;
  if (sign == op.sign)
  {
    if (CompareAbs (op) >= 0)
    {
      result.SubtractUnsigned (*this, op) ;
      result.sign = sign ;
    }
    else
    {
      result.SubtractUnsigned (op, *this) ;
      result.sign = - sign ;
    }
  }
  else
  {
    result.AddUnsigned (op, *this) ;
    result.sign = sign ;
  }
  result.CheckState () ;
  return result ;
}

currency currency::operator*(const currency &op) const
{
  currency result ;
  result.MultiplyUnsigned (*this, op) ;
  if (sign == op.sign)
    result.sign = 1 ;
  else
    result.sign = -1 ;
  result.CheckState () ;
  return result ;
}

currency currency::operator/(const currency &op) const
{
  currency result ;
  result.DivideUnsigned (*this, op) ;
  if (sign == op.sign)
    result.sign = 1 ;
  else
    result.sign = -1 ;
  result.CheckState () ;
  return result ;
}
currency currency::operator+=(const currency &op)
{
  currency result = *this + op ;
  this->operator=(result) ;
  return *this ;
}

currency currency::operator-=(const currency &op)
{
  currency result = *this - op ;
  this->operator=(result) ;
  return *this ;
}

currency currency::operator*=(const currency &op)
{
  currency result = *this * op ;
  this->operator=(result) ;
  return *this ;
}

currency currency::operator/=(const currency &op) 
{
  currency result = *this / op ;
  this->operator=(result) ;
  return *this ;
}

//
//  Description:
//
//    Unsigned addition. This function adds two currency values without
//    regard to sign (ie treats both operands as positive) and stores
//    the result in the object.
//
//  Parameters:
//
//    op1, op2: The two operands
//
void currency::AddUnsigned (const currency &op1, const currency &op2)
{
  signed char reg = 0 ;

  for (unsigned int ii = 0 ; ii < currency::PRECISION ; ++ ii)
  {
    reg =  (op1.data [ii] + op2.data [ii] + reg) ;
    data [ii] =  (reg % 10) ;
    reg /= 10 ;
  }
  data [PRECISION] = reg ;

  return ;
}

//
//  Description:
//
//    Unsigned subtraction. This function subtracts two currency values
//    without regard to sign (ie treats both operands as positive) and
//    stores the result in the object.
//
//  Parameters:
//
//    op1, op2: The two operands
//
void currency::SubtractUnsigned (const currency &op1, const currency &op2)
{
  for (unsigned int ii = 0 ; ii < PRECISION ; ++ ii)
    data [ii] = op1.data [ii] ;

  for (unsigned int ii = 0 ; ii < PRECISION ; ++ ii)
  {
    while (data [ii] < op2.data [ii])
    {
      data [ii] += 10 ;
      -- data [ii + 1] ;
    }
    data [ii] -= op2.data [ii] ;
  }
  return ;
}

//
//  Description:
//
//    Unsigned multiplication. This function multiplies two currency
//    values without regard to sign (ie treats both operands as positive)
//    and stores the result in the object.
//
//  Parameters:
//
//    op1, op2: The two operands
//

void currency::MultiplyUnsigned (const currency &op1, const currency &op2)
{
  int reg = 5 ; // Initialized to 5 to round up.
  for (unsigned int ii = 0 ; ii < PRECISION ; ++ ii)
  {
    for (unsigned int jj = 0 ; jj < PRECISION ; ++ jj)
    {
      reg = (op1.data [ii] * op2.data [jj] + data [jj + ii] + reg) ;
      data [jj + ii] = (reg % 10) ;
      reg /= 10 ;
    }
    reg = 0 ;
  }

  ShiftRight (DECIMALPLACES) ;
  for (unsigned int ii = PRECISION ; ii < REGISTERSIZE ; ++ ii)
  {
    if (data [ii] != 0)
      throw std::overflow_error ("value too large") ;
  }
  return ;
}

//
//  Description:
//
//    This function performs unsigned decimal division on two currency
//    objects and stores the result in this object.
//
//  Parameters:
//
//    op1, op2: op1 / op2
//
void currency::DivideUnsigned (const currency &op1, const currency &op2)
{
  currency reg1  ;
  currency reg2  ;

  bool op1zero = true ;
  bool op2zero = true ;
  // Here we check for zero valued operands and multiply the first
  // operand by 10. The multiplication allows us to round correctly.
  for (unsigned int ii = 0 ; ii < PRECISION ; ++ ii)
  {
    reg1.data [ii + DECIMALPLACES + 1] = op1.data [ii] ;
    if (reg1.data [ii] != 0)
      op1zero = false ;
    reg2.data [ii] = op2.data [ii] ;
    if (op2.data [ii] != 0)
      op2zero = false ;
  }

  if (op2zero)
    throw std::range_error ("Division by zero") ;

  // Dividing zero by anything besides zero gives zero.
  if (op1zero)
    return ;

  // Here we begin the long division. Find the power of ten that
  // creates a value larger than the first operand.
  int shiftcount = 0 ;
  while (reg1.CompareAbs (reg2) >= 0)
  {
    reg2.ShiftLeft (1) ;
    ++ shiftcount ;
  }

  // Integer division produces a zero
  if (shiftcount == 0)
    return ;

  // Move to the first position that is less than the operand.
  -- shiftcount ;
  reg2.ShiftRight (1) ;

  // Outer loop - Position in the output register.
  while (shiftcount >= 0)
  {
    // Next loop - find out how many times the divisor goes into
    // operand.
    while (reg1.CompareAbs (reg2) >= 0)
    {
      // Inner Loop - Subtraction
      for (unsigned int ii = 0 ; ii < REGISTERSIZE ; ++ ii)
      {
        while (reg1.data [ii] < reg2.data [ii])
        {
          assert (ii + 1 < REGISTERSIZE) ;

          reg1.data [ii] += 10 ;
          -- reg1.data [ii + 1] ;
        }
        reg1.data [ii] -= reg2.data [ii] ;
      }
      ++ data [shiftcount] ;
    }

    -- shiftcount ;  // Move to the next position in long division.
    reg2.ShiftRight (1) ;
  }

  // Round up and divide by 10.
  int reg = 5 ;
  for (int ii = 0 ; ii < REGISTERSIZE ; ++ ii)
  {
    reg = data [ii] + reg ;
    data [ii] = reg % 10 ;
    reg = reg / 10 ;
  }
  ShiftRight (1) ;

  return ;
}

//
//  Description:
//
//    This function does a decimal right shift, in other words a
//    division by power of 10.
//
//  Parameters:
//
//    count: The number of positions to shift.
//
void currency::ShiftRight (unsigned int count)
{
  for (unsigned int ii = count, jj = 0 ; ii < REGISTERSIZE ; ++ ii, ++ jj)
    data [jj] = data [ii] ;
  for (unsigned int ii = REGISTERSIZE - count ; ii < REGISTERSIZE ; ++ ii)
    data [ii] = 0 ;
  return ;
}

//
//  Description:
//
//    This function does a decimal left shift, in other words a
//    multiplication by power of 10.
//
//  Parameters:
//
//    count: The number of positions to shift.
//
void currency::ShiftLeft (unsigned int count)
{
  for (int ii = REGISTERSIZE - 1, jj = REGISTERSIZE - 1 - count ;
       ii > count ;
       -- jj, -- ii)
  {
    data [ii] = data [jj] ;
  }
  for (unsigned int ii = 0 ; ii < count ; ++ ii)
    data [ii] = 0 ;
}

//
//  Description:
//
//    This function converts the value of a currency object to
//    a "check style" text representation.
//
//  Parameters:
//
//    The number of decimal places in the output.
//
//  Return Value:
//
//    The formatted string.
//

std::string currency::ToText (unsigned int decimalplaces)
{
  string result ;

  // Here we format from the least significant whole digit to
  // the most significan. We build this portion of the string
  // from right to left. We work with three digits at a time.
  for (unsigned int ii = DECIMALPLACES, jj = 0 ;
       jj < PRECISION - DECIMALPLACES ;
       jj += 3, ii += 3)
  {
    string group ;
    if (ii + 2 < PRECISION && data [ii + 2] != 0)
    {
      group = numbers [data [ii + 2]] ;
      group += " Hundred" ;
    }
    int index = 0 ;
    if (ii + 1 < PRECISION)
      index = 10 * data [ii + 1] ;
    index += data [ii] ;
    if (index > 0)
    {
      if (group.length () != 0 && index != 0)
        group += " and " ;

      group += numbers [index] ;
    }
    if (index != 0 || group.length ())
    {
      switch (jj)
      {
      case 0: break ;
      case 3: group += " Thousand" ; break ;
      case 6: group += " Million" ; break ;
      case 9: group += " Billion" ; break ;
      case 12: group += " Trillion" ; break ;
      default: result = "*** VALUE TOO LARGE ***" ; return result ;
      }
      if (result.length () != 0)
        result = group + " " + result ;
      else
        result = group ;
    }
  }
  // Format the fraction

  if (decimalplaces != 0)
  {
    int places = decimalplaces ;
    if (places > DECIMALPLACES)
      places = DECIMALPLACES ;
    result += " And " ;
    bool nonzero = false ;
    for (int ii = DECIMALPLACES - 1, jj = 0 ; jj < places ; -- ii, ++ jj)
    {
      if (data [ii] != 0 || nonzero)
      {
        result += (data [ii] + '0') ;
        nonzero = true ;
      }
    }
    if (! nonzero)
      result += "No" ;
    result += "/1" ;
    for (int ii = 0 ; ii < places ; ++ ii)
      result += '0' ;
    result += "ths" ;
  }
  return result ;
}
//
//  Description:
//
//    This function rounds a currency value to a fixed number of
//    decimal places.
//
//  Parameters:
//
//    places: The number of decimal places to round to. If this value
//            is greater than the number of decimal places in the object
//            no rounding takes place.
//
//  Return Value:
//
//    A rounded currency object
//
currency currency::Round (unsigned int places) const
{
  currency result ;
  if (places != DECIMALPLACES)
  {
    currency rounding (5) ;
    rounding.ShiftRight (places + 1) ;
    result.AddUnsigned (*this, rounding) ;
    result.sign = sign ;
  }
  else
  {
    result = *this ;
  }

  int start ;
  if (places > DECIMALPLACES)
    start = DECIMALPLACES ;
  else
    start = places ;
  for (unsigned int ii = 0 ; ii < start ; ++ ii)
    result.data [ii] = 0 ;

  bool allzero = true ;
  for (unsigned int ii = start ; ii < PRECISION ; ++ ii)
  {
    if (result.data [ii] != 0)
      allzero = false ;
  }
  if (allzero)
    result.sign = 1 ;

  return result ;
}

//
//  Description:
//
//    integer cast operator
//
//  Return Value:
//
//    The currency value rounded to the nearest integer. If you want the
//    fractions, you need to multiply by a power of 10.
//
currency::operator int()
{
  currency temp (*this) ;
  temp.Round (0) ;

  int result = 0 ;
  for (int ii = PRECISION - 1 ; ii >= DECIMALPLACES ; -- ii)
  {
    result *= 10 ;
    result += temp.data [ii] ;
  }
  result *= temp.sign ;
  return result ;
}

//
//  Description:
//
//    double cast operator
//
//  Return Value:
//
//    The object value converted to double with all decimal precision
//    available.
//
currency::operator double ()
{
  double result = 0 ;
  for (int ii = PRECISION - 1 ; ii >= 0 ; -- ii)
  {
    result *= 10.0 ;
    result += data [ii] ;
  }
  for (int ii = 0 ; ii < DECIMALPLACES ; ++ ii)
    result /= 10.0 ;
  result *= sign ;
  return result ;
}

//
//  Description:
//
//    This function checks the state of a currency object to ensure
//    that it has not been corrupted.
//
void currency::CheckState () const
{
  switch (sign)
  {
  case 1: case -1: break ;
  default:  throw logic_error ("currency sign corrupt") ;
  }
  for (unsigned int ii = 0 ; ii < PRECISION ; ++ ii)
  {
    if (data [ii] < 0 || data [ii] > 9)
      throw logic_error ("currency data corrupt") ;
  }
  return ;
}


//************************************************************************
//************************************************************************
//**                                                                    **
//**  currency Formatting Classes                                       **
//**                                                                    **
//************************************************************************
//************************************************************************

//
//  Description:
//
//    The currency formatting classes are used as operands to
//    the ostream << operator. Each has a constructor that takes
//    a formatting setting parameter and stores it in a member variable
//    and a virtual Update () method that stores the value into an
//    ostream.
//

currency::WholeDigits::WholeDigits (unsigned int vv) : value (vv) {}
void currency::WholeDigits::Update (ostream &strm) const
{
  if (strm.iword (currency::whole_index) == 0
      && value != 0)
  {
    strm.iword (currency::decimal_index) = DEFAULTDECIMALPLACES ;
  }
  if (value > currency::PRECISION)
    strm.iword (currency::whole_index) = currency::PRECISION ;
  else
    strm.iword (currency::whole_index) = value ;
  return ;
}

currency::DecimalPlaces::DecimalPlaces (unsigned int vv) : value (vv) {}
void currency::DecimalPlaces::Update (ostream &strm) const
{
  if (value > currency::DECIMALPLACES)
    strm.iword (currency::decimal_index) = currency::DECIMALPLACES ;
  else
    strm.iword (currency::decimal_index) = value ;
  return ;
}

currency::NegativeFormat::NegativeFormat (unsigned int vv) : value (vv) {}
void currency::NegativeFormat::Update (ostream &strm) const
{
  strm.iword (currency::negative_index) = value ;
  return ;
}

currency::FillCharacter::FillCharacter (unsigned int vv) : value (vv) {}
void currency::FillCharacter::Update (ostream &strm) const
{
  strm.iword (currency::fill_index) = value ;
  return ;
}

currency::SignAlways::SignAlways (bool vv) : value (vv) {}
void currency::SignAlways::Update (ostream &strm) const
{
  if (value)
    strm.iword (currency::flags_index) |= SIGN_FLAG ;
  else
    strm.iword (currency::flags_index) &= ~SIGN_FLAG ;
  return ;
}

currency::UseCommas::UseCommas (bool vv) : value (vv) {}
void currency::UseCommas::Update (ostream &strm) const
{
  if (value)
    strm.iword (currency::flags_index) |= COMMA_FLAG ;
  else
    strm.iword (currency::flags_index) &= ~COMMA_FLAG ;
}

//************************************************************************
//************************************************************************
//**                                                                    **
//**  End Currency Formatting Classes                                   **
//**                                                                    **
//************************************************************************
//************************************************************************

