#ifndef __CURRENCY_H
#define __CURRENCY_H
//
//  Copyright 1999, Colosseum Builders, Inc.
//  info@colosseumbuilders.com
//
//  Title: Currency Class
//
//  Version:   1  Februrary 1st, 1999
//
//  Author:  John M. Miano miano@colosseumbuilders.com
//
//  Description:
//
//    This is a decimal, fixed-point currency class. Every operation is
//    carried out to currency::PRECISION digits of precision with
//    currency::DECIMALPLACES decimal places. It is designed to meet the
//    special mathematical needs of currency. Unlike a double,
//    this class represents fractions (pennies) exactly and rounds
//    in a predictable manner.
//
//    Although this is configurable, it is set up to represent values in the
//    range  -999,999,999,999.999 to 999,999,999,999.999 exactly.
//
//    currency::DECIMALPLACES is the number of decimal places retained
//    during mathematical calculations. The default number of decimal
//    places output is 2 (pennies). In other words, it is set up to
//    retain one extra digit behind the scenes.
//
//
//    Creating a currency Object:
//
//       currency (int)
//       currency (double)
//       currency (string) where string is in the format
//                         [-+]{digit}*([\.]?{digit}*}?
//                         leading and trailing spaces are allowed.
//                         Bad strings create a value of zero. Overflows
//                         raise overflow_error.
//
//    Arithmetic and comparison.
//
//      All arithmetic and comparison operations are performed with
//      currency operands. Other types are converted to currency
//      before the operation.
//      +,-,*,/,+=,-=,*=,/=,>,<,<=,>=,==,!= work as normal.
//
//      Assignment is defined for currency objects.
//
//    I/O
//
//      The << operator is defined for currency.
//
//      The following manipulators are also defined:
//
//      strm << currency::WholeDigits (n)
//      Sets the number of whole digits output (left of decimal) to n.
//
//      strm << currency::DecimalPlaces (n)
//      Sets the number of fractional digits output to n.
//
//      strm << currency::NegativeFormat (x) ;
//      Controls the representation of negative numbers.
//      x can be:
//        currency::leading   =>     -12345.67
//        currency::trailing  =>      12345.67-
//        currency::paren     =>     (12345.67)
//
//      strm << currency::FillCharacter (c)
//      Sets the fill character to c. This is used
//      in conjunction with WholeDigits. If more whole digits
//      have been requested than are needed to output (eg. WholeDigits (10)
//      when the currency value is 1234) the fill character is used to
//      pad the output.  A fill character of '0' uses inserts leading
//      zeros. A zero or space for the fill character means no fill is
//      used.
//
//      strm << currency::SignAlways (bool)
//      Determines if a + sign is output with positive numbers.
//      This has no effect when the Negative Format is "paren".
//
//      strm << currency::UseCommas (bool)
//      Determines if commas are used to group digits (e.g. 1,234,567.99).
//
//      strm.width (n) / strm << setw (n) is the only standard
//      manipulator used by the currency class.
//
//      IMPORTANT: If a number cannot be correctly output because either
//      the Whole Digits or width/setw setting is too small, the digits
//      are replaced with a '#' character to let the user know there
//      was an error.
//
//  String Conversion:
//
//    The ToText() member function converts a currency value to a text
//    string as used on checks. (e.g. 123.45 becomes "One Hundred and
//    Twenty Three and 45/100ths".)
//

class currency
{
private:
  class Register ;
  friend class currency::Register ;
public:
  currency (int value = 0) ;
  currency (double) ;
  currency (const currency &) ;
  currency &operator=(const currency &) ;
  currency (const std::string &) ;

  operator int() ;
  operator double () ;

  void Print (std::ostream &) const ;
  std::string ToText (unsigned int decimalplaces = 2) ;

  currency operator-() const ;
  currency operator+(const currency &) const ;
  currency operator-(const currency &) const ;
  currency operator*(const currency &) const ;
  currency operator/(const currency &) const ;

  currency operator+=(const currency &) ;
  currency operator-=(const currency &) ;
  currency operator*=(const currency &) ;
  currency operator/=(const currency &) ;

  bool operator<(const currency &) const ;
  bool operator<=(const currency &) const ;
  bool operator>(const currency &) const ;
  bool operator>=(const currency &) const ;
  bool operator==(const currency &) const ;
  bool operator!=(const currency &) const ;

  enum { leading = 0, trailing, paren,  } ;
  class CurrencyFormat ;
  class WholeDigits ;
  class DecimalPlaces ;
  class NegativeFormat ;
  class FillCharacter ;
  class UseCommas ;
  class SignAlways ;

  friend class currency::WholeDigits ;
  friend class currency::DecimalPlaces ;
  friend class currency::NegativeFormat ;
  friend class currency::FillCharacter ;
  friend class currency::SignAlways ;
  friend class currency::UseCommas ;
  friend class CurrencyInitialization ;

protected:
  enum { DECIMALPLACES = 3, } ;
  enum { WHOLEDIGITS = 15, } ;
  enum { PRECISION = DECIMALPLACES + WHOLEDIGITS, } ;

  void MultiplyUnsigned (const currency &, const currency &) ;
  void AddUnsigned (const currency &, const currency &) ;
  void SubtractUnsigned (const currency &, const currency &) ;
  void DivideUnsigned (const currency &, const currency &) ;

  void ShiftRight (unsigned int) ;
  void ShiftLeft (unsigned int) ;

  int Compare (const currency &) const ;
  int CompareAbs (const currency &) const ;

  currency Round (unsigned int places) const ;

  void CheckState () const ;

private:
  void DoCopy (const currency &) ;

  enum { REGISTERSIZE = 2 * PRECISION + 1, } ;
  typedef signed char REGTYPE ;
  REGTYPE data [REGISTERSIZE] ;
  int sign ;

  static int whole_index ;
  static int decimal_index ;
  static int negative_index ;
  static int fill_index ;
  static int flags_index ;
} ;



inline std::ostream &operator<<(std::ostream &strm, const currency &op)
{
  op.Print (strm) ;
  return strm ;
}

class currency::CurrencyFormat
{
public:
  virtual void Update (std::ostream &) const = 0 ;
protected:
  currencyFormat (){}
} ;

class currency::WholeDigits : public currency::CurrencyFormat
{
public:
  WholeDigits (unsigned int) ;
  virtual void Update (std::ostream &) const ;
private:
  unsigned int value ;
} ;

class currency::DecimalPlaces : public currency::CurrencyFormat
{
public:
  DecimalPlaces (unsigned int) ;
  virtual void Update (std::ostream &) const ;
private:
  unsigned int value ;
} ;

class currency::NegativeFormat :  public currency::CurrencyFormat
{
public:
  NegativeFormat (unsigned int) ;
  virtual void Update (std::ostream &) const ;
private:
  unsigned int value ;
} ;

class currency::FillCharacter :  public currency::CurrencyFormat
{
public:
  FillCharacter (unsigned int) ;
  virtual void Update (std::ostream &) const ;
private:
  unsigned int value ;
} ;

class currency::UseCommas :  public currency::CurrencyFormat
{
public:
  UseCommas (bool) ;
  virtual void Update (std::ostream &) const ;
private:
  bool value ;
} ;

class currency::SignAlways :  public currency::CurrencyFormat
{
public:
  SignAlways (bool) ;
  virtual void Update (std::ostream &) const ;
private:
  bool value ;
} ;

inline std::ostream &operator<<(std::ostream &strm,
                                const currency::CurrencyFormat &fmt)
{
  fmt.Update (strm) ;
  return strm ;
}


#endif
